# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2012 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Link Layer Discovery Protocol(LLDP, IEEE 802.1AB)
http://standards.ieee.org/getieee802/download/802.1AB-2009.pdf


basic TLV format::

    octets | 1          | 2             | 3 ...             n + 2 |
           --------------------------------------------------------
           | TLV type | TLV information | TLV information string  |
           | (7bits)  | string length   | (0-507 octets)          |
           |          | (9bits)         |                         |
           --------------------------------------------------------
    bits   |8        2|1|8             1|


Organizationally specific TLV format::

    octets | 1          | 2        | 3 ...  5 | 6       | 7 ...    n + 6 |
           ---------------------------------------------------------------
           | TLV type | Length     | OUI      | Subtype | Infomation     |
           | (7bits)  | (9bits)    | (24bits) | (8bits) | (0-507 octets) |
           ---------------------------------------------------------------
    bits   |8        2|1|8        1|


LLDPDU format::

    ------------------------------------------------------------------------
    | Chassis ID | Port ID | TTL | optional TLV | ... | optional TLV | End |
    ------------------------------------------------------------------------

Chasis ID, Port ID, TTL, End are mandatory
optional TLV may be inserted in any order
"""

import struct
from os_ken.lib import stringify
from os_ken.lib.packet import packet_base


# LLDP destination MAC address
LLDP_MAC_NEAREST_BRIDGE = '01:80:c2:00:00:0e'
LLDP_MAC_NEAREST_NON_TPMR_BRIDGE = '01:80:c2:00:00:03'
LLDP_MAC_NEAREST_CUSTOMER_BRIDGE = '01:80:c2:00:00:00'


LLDP_TLV_TYPELEN_STR = '!H'
LLDP_TLV_SIZE = 2
LLDP_TLV_TYPE_MASK = 0xfe00
LLDP_TLV_TYPE_SHIFT = 9
LLDP_TLV_LENGTH_MASK = 0x01ff


# LLDP TLV type
LLDP_TLV_END = 0                        # End of LLDPDU
LLDP_TLV_CHASSIS_ID = 1                 # Chassis ID
LLDP_TLV_PORT_ID = 2                    # Port ID
LLDP_TLV_TTL = 3                        # Time To Live
LLDP_TLV_PORT_DESCRIPTION = 4           # Port Description
LLDP_TLV_SYSTEM_NAME = 5                # System Name
LLDP_TLV_SYSTEM_DESCRIPTION = 6         # System Description
LLDP_TLV_SYSTEM_CAPABILITIES = 7        # System Capabilities
LLDP_TLV_MANAGEMENT_ADDRESS = 8         # Management Address
LLDP_TLV_ORGANIZATIONALLY_SPECIFIC = 127  # organizationally Specific TLVs


class LLDPBasicTLV(stringify.StringifyMixin):
    _LEN_MIN = 0
    _LEN_MAX = 511
    tlv_type = None

    def __init__(self, buf=None, *_args, **_kwargs):
        super(LLDPBasicTLV, self).__init__()
        if buf:
            (self.typelen, ) = struct.unpack(
                LLDP_TLV_TYPELEN_STR, buf[:LLDP_TLV_SIZE])
            tlv_type = \
                (self.typelen & LLDP_TLV_TYPE_MASK) >> LLDP_TLV_TYPE_SHIFT
            assert self.tlv_type == tlv_type

            self.len = self.typelen & LLDP_TLV_LENGTH_MASK
            assert len(buf) >= self.len + LLDP_TLV_SIZE

            self.tlv_info = buf[LLDP_TLV_SIZE:]
            self.tlv_info = self.tlv_info[:self.len]

    @staticmethod
    def get_type(buf):
        (typelen, ) = struct.unpack(LLDP_TLV_TYPELEN_STR, buf[:LLDP_TLV_SIZE])
        return (typelen & LLDP_TLV_TYPE_MASK) >> LLDP_TLV_TYPE_SHIFT

    @staticmethod
    def set_tlv_type(subcls, tlv_type):
        assert issubclass(subcls, LLDPBasicTLV)
        subcls.tlv_type = tlv_type

    def _len_valid(self):
        return self._LEN_MIN <= self.len and self.len <= self._LEN_MAX


class lldp(packet_base.PacketBase):
    """LLDPDU encoder/decoder class.

    An instance has the following attributes at least.

    ============== =====================================
    Attribute      Description
    ============== =====================================
    tlvs           List of TLV instance.
    ============== =====================================
    """
    _tlv_parsers = {}

    def __init__(self, tlvs):
        super(lldp, self).__init__()
        self.tlvs = tlvs

    # at least it must have chassis id, port id, ttl and end
    def _tlvs_len_valid(self):
        return len(self.tlvs) >= 4

    # chassis id, port id, ttl and end
    def _tlvs_valid(self):
        return (self.tlvs[0].tlv_type == LLDP_TLV_CHASSIS_ID and
                self.tlvs[1].tlv_type == LLDP_TLV_PORT_ID and
                self.tlvs[2].tlv_type == LLDP_TLV_TTL and
                self.tlvs[-1].tlv_type == LLDP_TLV_END)

    @classmethod
    def _parser(cls, buf):
        tlvs = []

        while buf:
            tlv_type = LLDPBasicTLV.get_type(buf)
            tlv = cls._tlv_parsers[tlv_type](buf)
            tlvs.append(tlv)
            offset = LLDP_TLV_SIZE + tlv.len
            buf = buf[offset:]
            if tlv.tlv_type == LLDP_TLV_END:
                break
            assert len(buf) > 0

        lldp_pkt = cls(tlvs)

        assert lldp_pkt._tlvs_len_valid()
        assert lldp_pkt._tlvs_valid()

        return lldp_pkt, None, buf

    @classmethod
    def parser(cls, buf):
        try:
            return cls._parser(buf)
        except:
            return None, None, buf

    def serialize(self, payload, prev):
        data = bytearray()
        for tlv in self.tlvs:
            data += tlv.serialize()

        return data

    @classmethod
    def set_type(cls, tlv_cls):
        cls._tlv_parsers[tlv_cls.tlv_type] = tlv_cls

    @classmethod
    def get_type(cls, tlv_type):
        return cls._tlv_parsers[tlv_type]

    @classmethod
    def set_tlv_type(cls, tlv_type):
        def _set_type(tlv_cls):
            tlv_cls.set_tlv_type(tlv_cls, tlv_type)
            cls.set_type(tlv_cls)
            return tlv_cls
        return _set_type

    def __len__(self):
        return sum(LLDP_TLV_SIZE + tlv.len for tlv in self.tlvs)


@lldp.set_tlv_type(LLDP_TLV_END)
class End(LLDPBasicTLV):
    """End TLV encoder/decoder class

    ============== =====================================
    Attribute      Description
    ============== =====================================
    buf            Binary data to parse.
    ============== =====================================
    """

    def __init__(self, buf=None, *args, **kwargs):
        super(End, self).__init__(buf, *args, **kwargs)
        if buf:
            pass
        else:
            self.len = 0
            self.typelen = 0

    def serialize(self):
        return struct.pack('!H', self.typelen)


@lldp.set_tlv_type(LLDP_TLV_CHASSIS_ID)
class ChassisID(LLDPBasicTLV):
    """Chassis ID TLV encoder/decoder class

    ============== =====================================
    Attribute      Description
    ============== =====================================
    buf            Binary data to parse.
    subtype        Subtype.
    chassis_id     Chassis id corresponding to subtype.
    ============== =====================================
    """

    _PACK_STR = '!B'
    _PACK_SIZE = struct.calcsize(_PACK_STR)
    # subtype id(1 octet) + chassis id length(1 - 255 octet)
    _LEN_MIN = 2
    _LEN_MAX = 256

    # Chassis ID subtype
    SUB_CHASSIS_COMPONENT = 1   # EntPhysicalAlias (IETF RFC 4133)
    SUB_INTERFACE_ALIAS = 2     # IfAlias (IETF RFC 2863)
    SUB_PORT_COMPONENT = 3      # EntPhysicalAlias (IETF RFC 4133)
    SUB_MAC_ADDRESS = 4         # MAC address (IEEE std 802)
    SUB_NETWORK_ADDRESS = 5     # networkAddress
    SUB_INTERFACE_NAME = 6      # IfName (IETF RFC 2863)
    SUB_LOCALLY_ASSIGNED = 7    # local

    def __init__(self, buf=None, *args, **kwargs):
        super(ChassisID, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.subtype, ) = struct.unpack(
                self._PACK_STR, self.tlv_info[:self._PACK_SIZE])
            self.chassis_id = self.tlv_info[self._PACK_SIZE:]
        else:
            self.subtype = kwargs['subtype']
            self.chassis_id = kwargs['chassis_id']
            self.len = self._PACK_SIZE + len(self.chassis_id)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!HB', self.typelen, self.subtype) + self.chassis_id


@lldp.set_tlv_type(LLDP_TLV_PORT_ID)
class PortID(LLDPBasicTLV):
    """Port ID TLV encoder/decoder class

    ============== =====================================
    Attribute      Description
    ============== =====================================
    buf            Binary data to parse.
    subtype        Subtype.
    port_id        Port ID corresponding to subtype.
    ============== =====================================
    """
    _PACK_STR = '!B'
    _PACK_SIZE = struct.calcsize(_PACK_STR)

    # subtype id(1 octet) + port id length(1 - 255 octet)
    _LEN_MIN = 2
    _LEN_MAX = 256

    # Port ID subtype
    SUB_INTERFACE_ALIAS = 1     # ifAlias (IETF RFC 2863)
    SUB_PORT_COMPONENT = 2      # entPhysicalAlias (IETF RFC 4133)
    SUB_MAC_ADDRESS = 3         # MAC address (IEEE Std 802)
    SUB_NETWORK_ADDRESS = 4     # networkAddress
    SUB_INTERFACE_NAME = 5      # ifName (IETF RFC 2863)
    SUB_AGENT_CIRCUIT_ID = 6    # agent circuit ID(IETF RFC 3046)
    SUB_LOCALLY_ASSIGNED = 7    # local

    def __init__(self, buf=None, *args, **kwargs):
        super(PortID, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.subtype, ) = struct.unpack(
                self._PACK_STR, self.tlv_info[:self._PACK_SIZE])
            self.port_id = self.tlv_info[self._PACK_SIZE:]
        else:
            self.subtype = kwargs['subtype']
            self.port_id = kwargs['port_id']
            self.len = self._PACK_SIZE + len(self.port_id)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!HB', self.typelen, self.subtype) + self.port_id


@lldp.set_tlv_type(LLDP_TLV_TTL)
class TTL(LLDPBasicTLV):
    """Time To Live TLV encoder/decoder class

    ============== =====================================
    Attribute      Description
    ============== =====================================
    buf            Binary data to parse.
    ttl            Time To Live.
    ============== =====================================
    """
    _PACK_STR = '!H'
    _PACK_SIZE = struct.calcsize(_PACK_STR)
    _LEN_MIN = _PACK_SIZE
    _LEN_MAX = _PACK_SIZE

    def __init__(self, buf=None, *args, **kwargs):
        super(TTL, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.ttl, ) = struct.unpack(
                self._PACK_STR, self.tlv_info[:self._PACK_SIZE])
        else:
            self.ttl = kwargs['ttl']
            self.len = self._PACK_SIZE
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!HH', self.typelen, self.ttl)


@lldp.set_tlv_type(LLDP_TLV_PORT_DESCRIPTION)
class PortDescription(LLDPBasicTLV):
    """Port description TLV encoder/decoder class

    ================= =====================================
    Attribute         Description
    ================= =====================================
    buf               Binary data to parse.
    port_description  Port description.
    ================= =====================================
    """
    _LEN_MAX = 255

    def __init__(self, buf=None, *args, **kwargs):
        super(PortDescription, self).__init__(buf, *args, **kwargs)
        if buf:
            pass
        else:
            self.port_description = kwargs['port_description']
            self.len = len(self.port_description)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!H', self.typelen) + self.port_description

    @property
    def port_description(self):
        return self.tlv_info

    @port_description.setter
    def port_description(self, value):
        self.tlv_info = value


@lldp.set_tlv_type(LLDP_TLV_SYSTEM_NAME)
class SystemName(LLDPBasicTLV):
    """System name TLV encoder/decoder class

    ================= =====================================
    Attribute         Description
    ================= =====================================
    buf               Binary data to parse.
    system_name       System name.
    ================= =====================================
    """
    _LEN_MAX = 255

    def __init__(self, buf=None, *args, **kwargs):
        super(SystemName, self).__init__(buf, *args, **kwargs)
        if buf:
            pass
        else:
            self.system_name = kwargs['system_name']
            self.len = len(self.system_name)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!H', self.typelen) + self.tlv_info

    @property
    def system_name(self):
        return self.tlv_info

    @system_name.setter
    def system_name(self, value):
        self.tlv_info = value


@lldp.set_tlv_type(LLDP_TLV_SYSTEM_DESCRIPTION)
class SystemDescription(LLDPBasicTLV):
    """System description TLV encoder/decoder class

    =================== =====================================
    Attribute           Description
    =================== =====================================
    buf                 Binary data to parse.
    system_description  System description.
    =================== =====================================
    """
    _LEN_MAX = 255

    def __init__(self, buf=None, *args, **kwargs):
        super(SystemDescription, self).__init__(buf, *args, **kwargs)
        if buf:
            pass
        else:
            self.system_description = kwargs['system_description']
            self.len = len(self.system_description)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!H', self.typelen) + self.tlv_info

    @property
    def system_description(self):
        return self.tlv_info

    @system_description.setter
    def system_description(self, value):
        self.tlv_info = value


@lldp.set_tlv_type(LLDP_TLV_SYSTEM_CAPABILITIES)
class SystemCapabilities(LLDPBasicTLV):
    """System Capabilities TLV encoder/decoder class

    ================= =====================================
    Attribute         Description
    ================= =====================================
    buf               Binary data to parse.
    system_cap        System Capabilities.
    enabled_cap       Enabled Capabilities.
    ================= =====================================
    """
    # system cap(2) + enabled cap(2)
    _PACK_STR = '!HH'
    _PACK_SIZE = struct.calcsize(_PACK_STR)
    _LEN_MIN = _PACK_SIZE
    _LEN_MAX = _PACK_SIZE

    # System Capabilities
    CAP_REPEATER = (1 << 1)             # IETF RFC 2108
    CAP_MAC_BRIDGE = (1 << 2)           # IEEE Std 802.1D
    CAP_WLAN_ACCESS_POINT = (1 << 3)    # IEEE Std 802.11 MIB
    CAP_ROUTER = (1 << 4)               # IETF RFC 1812
    CAP_TELEPHONE = (1 << 5)            # IETF RFC 4293
    CAP_DOCSIS = (1 << 6)               # IETF RFC 4639 and IETF RFC 4546
    CAP_STATION_ONLY = (1 << 7)         # IETF RFC 4293
    CAP_CVLAN = (1 << 8)                # IEEE Std 802.1Q
    CAP_SVLAN = (1 << 9)                # IEEE Std 802.1Q
    CAP_TPMR = (1 << 10)                # IEEE Std 802.1Q

    def __init__(self, buf=None, *args, **kwargs):
        super(SystemCapabilities, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.system_cap, self.enabled_cap) = struct.unpack(
                self._PACK_STR, self.tlv_info[:self._PACK_SIZE])
        else:
            self.system_cap = kwargs['system_cap']
            self.enabled_cap = kwargs['enabled_cap']
            self.len = self._PACK_SIZE
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!HHH',
                           self.typelen, self.system_cap, self.enabled_cap)


@lldp.set_tlv_type(LLDP_TLV_MANAGEMENT_ADDRESS)
class ManagementAddress(LLDPBasicTLV):
    """Management Address TLV encoder/decoder class

    ================= =====================================
    Attribute         Description
    ================= =====================================
    buf               Binary data to parse.
    addr_subtype      Address type.
    addr              Device address.
    intf_subtype      Interface type.
    intf_num          Interface number.
    oid               Object ID.
    ================= =====================================
    """
    _LEN_MIN = 9
    _LEN_MAX = 167

    _ADDR_PACK_STR = '!BB'    # address string length, address subtype
    _ADDR_PACK_SIZE = struct.calcsize(_ADDR_PACK_STR)
    _ADDR_LEN_MIN = 1
    _ADDR_LEN_MAX = 31

    _INTF_PACK_STR = '!BIB'   # interface subtype, interface number, oid length
    _INTF_PACK_SIZE = struct.calcsize(_INTF_PACK_STR)
    _OID_LEN_MIN = 0
    _OID_LEN_MAX = 128

    def __init__(self, buf=None, *args, **kwargs):
        super(ManagementAddress, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.addr_len, self.addr_subtype) = struct.unpack(
                self._ADDR_PACK_STR, self.tlv_info[:self._ADDR_PACK_SIZE])
            assert self._addr_len_valid()
            offset = self._ADDR_PACK_SIZE + self.addr_len - 1
            self.addr = self.tlv_info[self._ADDR_PACK_SIZE:offset]

            (self.intf_subtype, self.intf_num, self.oid_len) = struct.unpack(
                self._INTF_PACK_STR,
                self.tlv_info[offset:offset + self._INTF_PACK_SIZE])
            assert self._oid_len_valid()

            offset = offset + self._INTF_PACK_SIZE
            self.oid = self.tlv_info[offset:]
        else:
            self.addr_subtype = kwargs['addr_subtype']
            self.addr = kwargs['addr']
            self.addr_len = len(self.addr) + 1  # 1 octet subtype
            assert self._addr_len_valid()

            self.intf_subtype = kwargs['intf_subtype']
            self.intf_num = kwargs['intf_num']

            self.oid = kwargs['oid']
            self.oid_len = len(self.oid)
            assert self._oid_len_valid()

            self.len = self._ADDR_PACK_SIZE + self.addr_len - 1 \
                + self._INTF_PACK_SIZE + self.oid_len
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        tlv_info = struct.pack(self._ADDR_PACK_STR,
                               self.addr_len, self.addr_subtype)
        tlv_info += self.addr
        tlv_info += struct.pack(self._INTF_PACK_STR,
                                self.intf_subtype, self.intf_num, self.oid_len)
        tlv_info += self.oid
        return struct.pack('!H', self.typelen) + tlv_info

    def _addr_len_valid(self):
        return (self._ADDR_LEN_MIN <= self.addr_len or
                self.addr_len <= self._ADDR_LEN_MAX)

    def _oid_len_valid(self):
        return self._OID_LEN_MIN <= self.oid_len <= self._OID_LEN_MAX


@lldp.set_tlv_type(LLDP_TLV_ORGANIZATIONALLY_SPECIFIC)
class OrganizationallySpecific(LLDPBasicTLV):
    """Organizationally Specific TLV encoder/decoder class

    ================= =============================================
    Attribute         Description
    ================= =============================================
    buf               Binary data to parse.
    oui               Organizationally unique ID.
    subtype           Organizationally defined subtype.
    info              Organizationally defined information string.
    ================= =============================================
    """
    _PACK_STR = '!3sB'
    _PACK_SIZE = struct.calcsize(_PACK_STR)
    _LEN_MIN = _PACK_SIZE
    _LEN_MAX = 511

    def __init__(self, buf=None, *args, **kwargs):
        super(OrganizationallySpecific, self).__init__(buf, *args, **kwargs)
        if buf:
            (self.oui, self.subtype) = struct.unpack(
                self._PACK_STR, self.tlv_info[:self._PACK_SIZE])
            self.info = self.tlv_info[self._PACK_SIZE:]
        else:
            self.oui = kwargs['oui']
            self.subtype = kwargs['subtype']
            self.info = kwargs['info']
            self.len = self._PACK_SIZE + len(self.info)
            assert self._len_valid()
            self.typelen = (self.tlv_type << LLDP_TLV_TYPE_SHIFT) | self.len

    def serialize(self):
        return struct.pack('!H3sB', self.typelen, self.oui,
                           self.subtype) + self.info


lldp.set_classes(lldp._tlv_parsers)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import unittest
import logging
import struct

from nose.tools import eq_
from os_ken.lib.packet import bpdu


LOG = logging.getLogger(__name__)


class Test_ConfigurationBPDUs(unittest.TestCase):
    """ Test case for ConfigurationBPDUs
    """

    def setUp(self):
        self.protocol_id = bpdu.PROTOCOL_IDENTIFIER
        self.version_id = bpdu.ConfigurationBPDUs.VERSION_ID
        self.bpdu_type = bpdu.ConfigurationBPDUs.BPDU_TYPE
        self.flags = 0b00000001
        self.root_priority = 4096
        self.root_system_id_extension = 1
        self.root_mac_address = '12:34:56:78:9a:bc'
        self.root_path_cost = 2
        self.bridge_priority = 8192
        self.bridge_system_id_extension = 3
        self.bridge_mac_address = 'aa:aa:aa:aa:aa:aa'
        self.port_priority = 16
        self.port_number = 4
        self.message_age = 5
        self.max_age = 6
        self.hello_time = 7
        self.forward_delay = 8

        self.msg = bpdu.ConfigurationBPDUs(
            flags=self.flags,
            root_priority=self.root_priority,
            root_system_id_extension=self.root_system_id_extension,
            root_mac_address=self.root_mac_address,
            root_path_cost=self.root_path_cost,
            bridge_priority=self.bridge_priority,
            bridge_system_id_extension=self.bridge_system_id_extension,
            bridge_mac_address=self.bridge_mac_address,
            port_priority=self.port_priority,
            port_number=self.port_number,
            message_age=self.message_age,
            max_age=self.max_age,
            hello_time=self.hello_time,
            forward_delay=self.forward_delay)

        self.fmt = (bpdu.bpdu._PACK_STR
                    + bpdu.ConfigurationBPDUs._PACK_STR[1:])
        self.buf = struct.pack(self.fmt,
                               self.protocol_id, self.version_id,
                               self.bpdu_type, self.flags,
                               bpdu.ConfigurationBPDUs.encode_bridge_id(
                                   self.root_priority,
                                   self.root_system_id_extension,
                                   self.root_mac_address),
                               self.root_path_cost,
                               bpdu.ConfigurationBPDUs.encode_bridge_id(
                                   self.bridge_priority,
                                   self.bridge_system_id_extension,
                                   self.bridge_mac_address),
                               bpdu.ConfigurationBPDUs.encode_port_id(
                                   self.port_priority,
                                   self.port_number),
                               bpdu.ConfigurationBPDUs._encode_timer(
                                   self.message_age),
                               bpdu.ConfigurationBPDUs._encode_timer(
                                   self.max_age),
                               bpdu.ConfigurationBPDUs._encode_timer(
                                   self.hello_time),
                               bpdu.ConfigurationBPDUs._encode_timer(
                                   self.forward_delay))

    def test_init(self):
        eq_(self.protocol_id, self.msg._protocol_id)
        eq_(self.version_id, self.msg._version_id)
        eq_(self.bpdu_type, self.msg._bpdu_type)
        eq_(self.flags, self.msg.flags)
        eq_(self.root_priority, self.msg.root_priority)
        eq_(self.root_system_id_extension,
            self.msg.root_system_id_extension)
        eq_(self.root_mac_address, self.msg.root_mac_address)
        eq_(self.root_path_cost, self.msg.root_path_cost)
        eq_(self.bridge_priority, self.msg.bridge_priority)
        eq_(self.bridge_system_id_extension,
            self.msg.bridge_system_id_extension)
        eq_(self.bridge_mac_address, self.msg.bridge_mac_address)
        eq_(self.port_priority, self.msg.port_priority)
        eq_(self.port_number, self.msg.port_number)
        eq_(self.message_age, self.msg.message_age)
        eq_(self.max_age, self.msg.max_age)
        eq_(self.hello_time, self.msg.hello_time)
        eq_(self.forward_delay, self.msg.forward_delay)

    def test_parser(self):
        r1, r2, _ = bpdu.bpdu.parser(self.buf)

        eq_(type(r1), type(self.msg))
        eq_(r1._protocol_id, self.protocol_id)
        eq_(r1._version_id, self.version_id)
        eq_(r1._bpdu_type, self.bpdu_type)
        eq_(r1.flags, self.flags)
        eq_(r1.root_priority, self.root_priority)
        eq_(r1.root_system_id_extension, self.root_system_id_extension)
        eq_(r1.root_mac_address, self.root_mac_address)
        eq_(r1.root_path_cost, self.root_path_cost)
        eq_(r1.bridge_priority, self.bridge_priority)
        eq_(r1.bridge_system_id_extension, self.bridge_system_id_extension)
        eq_(r1.bridge_mac_address, self.bridge_mac_address)
        eq_(r1.port_priority, self.port_priority)
        eq_(r1.port_number, self.port_number)
        eq_(r1.message_age, self.message_age)
        eq_(r1.max_age, self.max_age)
        eq_(r1.hello_time, self.hello_time)
        eq_(r1.forward_delay, self.forward_delay)
        eq_(r2, None)

    def test_serialize(self):
        data = bytearray()
        prev = None
        buf = self.msg.serialize(data, prev)
        res = struct.unpack(self.fmt, buf)

        eq_(res[0], self.protocol_id)
        eq_(res[1], self.version_id)
        eq_(res[2], self.bpdu_type)
        eq_(res[3], self.flags)
        eq_(bpdu.ConfigurationBPDUs._decode_bridge_id(res[4]),
            (self.root_priority,
             self.root_system_id_extension,
             self.root_mac_address))
        eq_(res[5], self.root_path_cost)
        eq_(bpdu.ConfigurationBPDUs._decode_bridge_id(res[6]),
            (self.bridge_priority,
             self.bridge_system_id_extension,
             self.bridge_mac_address))
        eq_(bpdu.ConfigurationBPDUs._decode_port_id(res[7]),
            (self.port_priority,
             self.port_number))
        eq_(bpdu.ConfigurationBPDUs._decode_timer(res[8]), self.message_age)
        eq_(bpdu.ConfigurationBPDUs._decode_timer(res[9]), self.max_age)
        eq_(bpdu.ConfigurationBPDUs._decode_timer(res[10]), self.hello_time)
        eq_(bpdu.ConfigurationBPDUs._decode_timer(res[11]), self.forward_delay)

    def test_json(self):
        jsondict = self.msg.to_jsondict()
        msg = bpdu.ConfigurationBPDUs.from_jsondict(
            jsondict['ConfigurationBPDUs'])
        eq_(str(self.msg), str(msg))


class Test_TopologyChangeNotificationBPDUs(unittest.TestCase):
    """ Test case for TopologyChangeNotificationBPDUs
    """

    def setUp(self):
        self.protocol_id = bpdu.PROTOCOL_IDENTIFIER
        self.version_id = bpdu.TopologyChangeNotificationBPDUs.VERSION_ID
        self.bpdu_type = bpdu.TopologyChangeNotificationBPDUs.BPDU_TYPE

        self.msg = bpdu.TopologyChangeNotificationBPDUs()

        self.fmt = bpdu.bpdu._PACK_STR
        self.buf = struct.pack(self.fmt,
                               self.protocol_id,
                               self.version_id,
                               self.bpdu_type)

    def test_init(self):
        eq_(self.protocol_id, self.msg._protocol_id)
        eq_(self.version_id, self.msg._version_id)
        eq_(self.bpdu_type, self.msg._bpdu_type)

    def test_parser(self):
        r1, r2, _ = bpdu.bpdu.parser(self.buf)

        eq_(type(r1), type(self.msg))
        eq_(r1._protocol_id, self.protocol_id)
        eq_(r1._version_id, self.version_id)
        eq_(r1._bpdu_type, self.bpdu_type)
        eq_(r2, None)

    def test_serialize(self):
        data = bytearray()
        prev = None
        buf = self.msg.serialize(data, prev)
        res = struct.unpack(self.fmt, buf)

        eq_(res[0], self.protocol_id)
        eq_(res[1], self.version_id)
        eq_(res[2], self.bpdu_type)

    def test_json(self):
        jsondict = self.msg.to_jsondict()
        msg = bpdu.TopologyChangeNotificationBPDUs.from_jsondict(
            jsondict['TopologyChangeNotificationBPDUs'])
        eq_(str(self.msg), str(msg))


class Test_RstBPDUs(unittest.TestCase):
    """ Test case for RstBPDUs
    """

    def setUp(self):
        self.protocol_id = bpdu.PROTOCOL_IDENTIFIER
        self.version_id = bpdu.RstBPDUs.VERSION_ID
        self.bpdu_type = bpdu.RstBPDUs.BPDU_TYPE
        self.flags = 0b01111110
        self.root_priority = 4096
        self.root_system_id_extension = 1
        self.root_mac_address = '12:34:56:78:9a:bc'
        self.root_path_cost = 2
        self.bridge_priority = 8192
        self.bridge_system_id_extension = 3
        self.bridge_mac_address = 'aa:aa:aa:aa:aa:aa'
        self.port_priority = 16
        self.port_number = 4
        self.message_age = 5
        self.max_age = 6
        self.hello_time = 7
        self.forward_delay = 8
        self.version_1_length = bpdu.VERSION_1_LENGTH

        self.msg = bpdu.RstBPDUs(
            flags=self.flags,
            root_priority=self.root_priority,
            root_system_id_extension=self.root_system_id_extension,
            root_mac_address=self.root_mac_address,
            root_path_cost=self.root_path_cost,
            bridge_priority=self.bridge_priority,
            bridge_system_id_extension=self.bridge_system_id_extension,
            bridge_mac_address=self.bridge_mac_address,
            port_priority=self.port_priority,
            port_number=self.port_number,
            message_age=self.message_age,
            max_age=self.max_age,
            hello_time=self.hello_time,
            forward_delay=self.forward_delay)

        self.fmt = (bpdu.bpdu._PACK_STR
                    + bpdu.ConfigurationBPDUs._PACK_STR[1:]
                    + bpdu.RstBPDUs._PACK_STR[1:])
        self.buf = struct.pack(self.fmt,
                               self.protocol_id, self.version_id,
                               self.bpdu_type, self.flags,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.root_priority,
                                   self.root_system_id_extension,
                                   self.root_mac_address),
                               self.root_path_cost,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.bridge_priority,
                                   self.bridge_system_id_extension,
                                   self.bridge_mac_address),
                               bpdu.RstBPDUs.encode_port_id(
                                   self.port_priority,
                                   self.port_number),
                               bpdu.RstBPDUs._encode_timer(self.message_age),
                               bpdu.RstBPDUs._encode_timer(self.max_age),
                               bpdu.RstBPDUs._encode_timer(self.hello_time),
                               bpdu.RstBPDUs._encode_timer(self.forward_delay),
                               self.version_1_length)

    def test_init(self):
        eq_(self.protocol_id, self.msg._protocol_id)
        eq_(self.version_id, self.msg._version_id)
        eq_(self.bpdu_type, self.msg._bpdu_type)
        eq_(self.flags, self.msg.flags)
        eq_(self.root_priority, self.msg.root_priority)
        eq_(self.root_system_id_extension,
            self.msg.root_system_id_extension)
        eq_(self.root_mac_address, self.msg.root_mac_address)
        eq_(self.root_path_cost, self.msg.root_path_cost)
        eq_(self.bridge_priority, self.msg.bridge_priority)
        eq_(self.bridge_system_id_extension,
            self.msg.bridge_system_id_extension)
        eq_(self.bridge_mac_address, self.msg.bridge_mac_address)
        eq_(self.port_priority, self.msg.port_priority)
        eq_(self.port_number, self.msg.port_number)
        eq_(self.message_age, self.msg.message_age)
        eq_(self.max_age, self.msg.max_age)
        eq_(self.hello_time, self.msg.hello_time)
        eq_(self.forward_delay, self.msg.forward_delay)
        eq_(self.version_1_length, self.msg._version_1_length)

    def test_parser(self):
        r1, r2, _ = bpdu.bpdu.parser(self.buf)

        eq_(type(r1), type(self.msg))
        eq_(r1._protocol_id, self.protocol_id)
        eq_(r1._version_id, self.version_id)
        eq_(r1._bpdu_type, self.bpdu_type)
        eq_(r1.flags, self.flags)
        eq_(r1.root_priority, self.root_priority)
        eq_(r1.root_system_id_extension, self.root_system_id_extension)
        eq_(r1.root_mac_address, self.root_mac_address)
        eq_(r1.root_path_cost, self.root_path_cost)
        eq_(r1.bridge_priority, self.bridge_priority)
        eq_(r1.bridge_system_id_extension, self.bridge_system_id_extension)
        eq_(r1.bridge_mac_address, self.bridge_mac_address)
        eq_(r1.port_priority, self.port_priority)
        eq_(r1.port_number, self.port_number)
        eq_(r1.message_age, self.message_age)
        eq_(r1.max_age, self.max_age)
        eq_(r1.hello_time, self.hello_time)
        eq_(r1.forward_delay, self.forward_delay)
        eq_(r1._version_1_length, self.version_1_length)
        eq_(r2, None)

    def test_serialize(self):
        data = bytearray()
        prev = None
        buf = self.msg.serialize(data, prev)
        res = struct.unpack(self.fmt, buf)

        eq_(res[0], self.protocol_id)
        eq_(res[1], self.version_id)
        eq_(res[2], self.bpdu_type)
        eq_(res[3], self.flags)
        eq_(bpdu.RstBPDUs._decode_bridge_id(res[4]),
            (self.root_priority,
             self.root_system_id_extension,
             self.root_mac_address))
        eq_(res[5], self.root_path_cost)
        eq_(bpdu.RstBPDUs._decode_bridge_id(res[6]),
            (self.bridge_priority,
             self.bridge_system_id_extension,
             self.bridge_mac_address))
        eq_(bpdu.RstBPDUs._decode_port_id(res[7]),
            (self.port_priority,
             self.port_number))
        eq_(bpdu.RstBPDUs._decode_timer(res[8]), self.message_age)
        eq_(bpdu.RstBPDUs._decode_timer(res[9]), self.max_age)
        eq_(bpdu.RstBPDUs._decode_timer(res[10]), self.hello_time)
        eq_(bpdu.RstBPDUs._decode_timer(res[11]), self.forward_delay)
        eq_(res[12], self.version_1_length)

    def test_json(self):
        jsondict = self.msg.to_jsondict()
        msg = bpdu.RstBPDUs.from_jsondict(jsondict['RstBPDUs'])
        eq_(str(self.msg), str(msg))


class Test_UnknownVersion(unittest.TestCase):
    """ Test case for unknown BPDU version
    """

    def setUp(self):
        self.protocol_id = bpdu.PROTOCOL_IDENTIFIER
        self.version_id = 111  # Unknown version
        self.bpdu_type = bpdu.RstBPDUs.BPDU_TYPE
        self.flags = 0b01111110
        self.root_priority = 4096
        self.root_system_id_extension = 1
        self.root_mac_address = '12:34:56:78:9a:bc'
        self.root_path_cost = 2
        self.bridge_priority = 8192
        self.bridge_system_id_extension = 3
        self.bridge_mac_address = 'aa:aa:aa:aa:aa:aa'
        self.port_priority = 16
        self.port_number = 4
        self.message_age = 5
        self.max_age = 6
        self.hello_time = 7
        self.forward_delay = 8
        self.version_1_length = bpdu.VERSION_1_LENGTH

        self.fmt = (bpdu.bpdu._PACK_STR
                    + bpdu.ConfigurationBPDUs._PACK_STR[1:]
                    + bpdu.RstBPDUs._PACK_STR[1:])
        self.buf = struct.pack(self.fmt,
                               self.protocol_id, self.version_id,
                               self.bpdu_type, self.flags,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.root_priority,
                                   self.root_system_id_extension,
                                   self.root_mac_address),
                               self.root_path_cost,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.bridge_priority,
                                   self.bridge_system_id_extension,
                                   self.bridge_mac_address),
                               bpdu.RstBPDUs.encode_port_id(
                                   self.port_priority,
                                   self.port_number),
                               bpdu.RstBPDUs._encode_timer(self.message_age),
                               bpdu.RstBPDUs._encode_timer(self.max_age),
                               bpdu.RstBPDUs._encode_timer(self.hello_time),
                               bpdu.RstBPDUs._encode_timer(self.forward_delay),
                               self.version_1_length)

    def test_parser(self):
        r1, r2, _ = bpdu.bpdu.parser(self.buf)
        eq_(r1, self.buf)
        eq_(r2, None)


class Test_UnknownType(unittest.TestCase):
    """ Test case for unknown BPDU type
    """

    def setUp(self):
        self.protocol_id = bpdu.PROTOCOL_IDENTIFIER
        self.version_id = bpdu.RstBPDUs.VERSION_ID
        self.bpdu_type = 222  # Unknown type
        self.flags = 0b01111110
        self.root_priority = 4096
        self.root_system_id_extension = 1
        self.root_mac_address = '12:34:56:78:9a:bc'
        self.root_path_cost = 2
        self.bridge_priority = 8192
        self.bridge_system_id_extension = 3
        self.bridge_mac_address = 'aa:aa:aa:aa:aa:aa'
        self.port_priority = 16
        self.port_number = 4
        self.message_age = 5
        self.max_age = 6
        self.hello_time = 7
        self.forward_delay = 8
        self.version_1_length = bpdu.VERSION_1_LENGTH

        self.fmt = (bpdu.bpdu._PACK_STR
                    + bpdu.ConfigurationBPDUs._PACK_STR[1:]
                    + bpdu.RstBPDUs._PACK_STR[1:])
        self.buf = struct.pack(self.fmt,
                               self.protocol_id, self.version_id,
                               self.bpdu_type, self.flags,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.root_priority,
                                   self.root_system_id_extension,
                                   self.root_mac_address),
                               self.root_path_cost,
                               bpdu.RstBPDUs.encode_bridge_id(
                                   self.bridge_priority,
                                   self.bridge_system_id_extension,
                                   self.bridge_mac_address),
                               bpdu.RstBPDUs.encode_port_id(
                                   self.port_priority,
                                   self.port_number),
                               bpdu.RstBPDUs._encode_timer(self.message_age),
                               bpdu.RstBPDUs._encode_timer(self.max_age),
                               bpdu.RstBPDUs._encode_timer(self.hello_time),
                               bpdu.RstBPDUs._encode_timer(self.forward_delay),
                               self.version_1_length)

    def test_parser(self):
        r1, r2, _ = bpdu.bpdu.parser(self.buf)
        eq_(r1, self.buf)
        eq_(r2, None)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE: this file is intendend to be replaced with mechanically generated
# file if/when OF-Config yang specification is available with a suitable
# license.

from os_ken.lib.of_config.base import _Base, _e, _ct


class OFPortConfigurationType(_Base):
    _ELEMENTS = [
        _e('admin-state', is_list=False),
        _e('no-receive', is_list=False),
        _e('no-forward', is_list=False),
        _e('no-packet-in', is_list=False),
    ]


class OFPortStateType(_Base):
    _ELEMENTS = [
        _e('oper-state', is_list=False),
        _e('blocked', is_list=False),
        _e('live', is_list=False),
    ]


class OFPortType(_Base):
    _ELEMENTS = [
        _e('resource-id', is_list=False),
        _e('number', is_list=False),
        _e('name', is_list=False),
        _e('current-rate', is_list=False),
        _e('max-rate', is_list=False),
        _ct('configuration', OFPortConfigurationType, is_list=False),
        _ct('state', OFPortStateType, is_list=False),
        _ct('features', None, is_list=False),
        _ct('tunnel-type', None, is_list=False),
    ]


class OFQueuePropertiesType(_Base):
    _ELEMENTS = [
        _e('min-rate', is_list=False),
        _e('max-rate', is_list=False),
        _e('experimenter', is_list=True),
    ]


class OFQueueType(_Base):
    _ELEMENTS = [
        _e('resource-id', is_list=False),
        _e('id', is_list=False),
        _e('port', is_list=False),
        _ct('properties', OFQueuePropertiesType, is_list=False),
    ]


class OFCapableSwitchResourcesType(_Base):
    _ELEMENTS = [
        _ct('port', OFPortType, is_list=True),
        _ct('queue', OFQueueType, is_list=True),
        _ct('owned-certificate', None, is_list=True),
        _ct('external-certificate', None, is_list=True),
        _ct('flow-table', None, is_list=True),
    ]


class OFControllerStateType(_Base):
    _ELEMENTS = [
        _e('connection-state', is_list=False),
        _e('current-version', is_list=False),

        # XXX OF-Config 1.1.1 is inconsistent about supported-versions.
        #
        # according to its xml schema (p.43), i believe this should look
        # like the following.  it's what linc/of_config does, too.
        #     <supported-versions>1.3</supported-versions>
        #
        # on the other hand, it has an example (p.45) like the following.
        # this one is compatible with OF-Config 1.1.
        #     <supported-versions>
        #         <version>1.2</version>
        #         <version>1.1</version>
        #     </supported-versions>

        _e('supported-versions', is_list=True),
        _e('local-ip-address-in-use', is_list=False),
        _e('local-port-in-use', is_list=False),
    ]


class OFControllerType(_Base):
    _ELEMENTS = [
        _e('id', is_list=False),
        _e('role', is_list=False),
        _e('ip-address', is_list=False),
        _e('port', is_list=False),
        _e('local-ip-address', is_list=False),
        _e('local-port', is_list=False),
        _e('protocol', is_list=False),
        _ct('state', OFControllerStateType, is_list=False),
    ]


class OFLogicalSwitchControllersType(_Base):
    _ELEMENTS = [
        _ct('controller', OFControllerType, is_list=True),
    ]


class OFLogicalSwitchResourcesType(_Base):
    _ELEMENTS = [
        _e('port', is_list=True),
        _e('queue', is_list=True),
        _e('certificate', is_list=False),
        _e('flow-table', is_list=True),
    ]


class OFLogicalSwitchType(_Base):
    _ELEMENTS = [
        _e('id', is_list=False),
        _ct('capabilities', None, is_list=False),
        _e('datapath-id', is_list=False),
        _e('enabled', is_list=False),
        _e('check-controller-certificate', is_list=False),
        _e('lost-connection-behavior', is_list=False),
        _ct('controllers', OFLogicalSwitchControllersType, is_list=False),
        _ct('resources', OFLogicalSwitchResourcesType, is_list=False),
    ]


class OFCapableSwitchLogicalSwitchesType(_Base):
    _ELEMENTS = [
        # this is named 'logical-switch' for OF-Config 1.1.
        _ct('switch', OFLogicalSwitchType, is_list=True),
    ]


class OFCapableSwitchType(_Base):
    _ELEMENTS = [
        _e('id', is_list=False),
        _e('config-version', is_list=False),
        _ct('configuration-points', None, is_list=False),
        _ct('resources', OFCapableSwitchResourcesType, is_list=False),
        _ct('logical-switches', OFCapableSwitchLogicalSwitchesType,
            is_list=False),
    ]



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import logging

from os_ken.ofproto import ofproto_v1_2
from os_ken.ofproto import ether
from os_ken.ofproto import inet
from os_ken.tests.integrated import tester

LOG = logging.getLogger(__name__)


class RunTest(tester.TestFlowBase):
    """ Test case for add flows of Matches
    """
    OFP_VERSIONS = [ofproto_v1_2.OFP_VERSION]

    def __init__(self, *args, **kwargs):
        super(RunTest, self).__init__(*args, **kwargs)

        self._verify = {}

    def add_matches(self, dp, match):
        m = dp.ofproto_parser.OFPFlowMod(dp, 0, 0, 0,
                                         dp.ofproto.OFPFC_ADD,
                                         0, 0, 0, 0xffffffff,
                                         dp.ofproto.OFPP_ANY,
                                         0xffffffff, 0, match, [])
        dp.send_msg(m)

    def _set_verify(self, headers, value, mask=None,
                    all_bits_masked=False, type_='int'):
        self._verify = {}
        self._verify['headers'] = headers
        self._verify['value'] = value
        self._verify['mask'] = mask
        self._verify['all_bits_masked'] = all_bits_masked
        self._verify['type'] = type_

    def verify_default(self, dp, stats):
        type_ = self._verify['type']
        headers = self._verify['headers']
        value = self._verify['value']
        mask = self._verify['mask']
        value_masked = self._masked(type_, value, mask)
        all_bits_masked = self._verify['all_bits_masked']

        field = None
        for s in stats:
            for f in s.match.fields:
                if f.header in headers:
                    field = f
                    break

        if field is None:
            if self._is_all_zero_bit(type_, mask):
                return True
            return 'Field not found.'

        f_value = field.value
        if hasattr(field, 'mask'):
            f_mask = field.mask
        else:
            f_mask = None

        if (f_value == value) or (f_value == value_masked):
            if (f_mask == mask) or (all_bits_masked and f_mask is None):
                return True

        return "send: %s/%s, reply: %s/%s" \
            % (self._cnv_to_str(type_, value, mask, f_value, f_mask))

    def _masked(self, type_, value, mask):
        if mask is None:
            v = value
        elif type_ == 'int':
            v = value & mask
        elif type_ == 'mac':
            v = self.haddr_masked(value, mask)
        elif type_ == 'ipv4':
            v = self.ipv4_masked(value, mask)
        elif type_ == 'ipv6':
            v = self.ipv6_masked(value, mask)
        else:
            raise Exception('Unknown type')
        return v

    def _is_all_zero_bit(self, type_, val):
        if type_ == 'int' or type_ == 'ipv4':
            return val == 0
        elif type_ == 'mac':
            for v in val:
                if v != b'\x00':
                    return False
            return True
        elif type_ == 'ipv6':
            for v in val:
                if v != 0:
                    return False
            return True
        else:
            raise Exception('Unknown type')

    def _cnv_to_str(self, type_, value, mask, f_value, f_mask):
        func = None
        if type_ == 'int':
            pass
        elif type_ == 'mac':
            func = self.haddr_to_str
        elif type_ == 'ipv4':
            func = self.ipv4_to_str
        elif type_ == 'ipv6':
            func = self.ipv6_to_str
        else:
            raise Exception('Unknown type')

        if func:
            value = func(value)
            f_value = func(f_value)
            if mask:
                mask = func(mask)
            if f_mask:
                f_mask = func(f_mask)

        return value, mask, f_value, f_mask

    def test_rule_set_dl_dst(self, dp):
        dl_dst = 'e2:7a:09:79:0b:0f'
        dl_dst_bin = self.haddr_to_bin(dl_dst)

        headers = [dp.ofproto.OXM_OF_ETH_DST, dp.ofproto.OXM_OF_ETH_DST_W]
        self._set_verify(headers, dl_dst_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_dst(dl_dst_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_dst_masked_ff(self, dp):
        dl_dst = 'd0:98:79:b4:75:b5'
        dl_dst_bin = self.haddr_to_bin(dl_dst)
        mask = 'ff:ff:ff:ff:ff:ff'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_DST, dp.ofproto.OXM_OF_ETH_DST_W]
        self._set_verify(headers, dl_dst_bin, mask_bin, True, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_dst_masked(dl_dst_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_dst_masked_f0(self, dp):
        dl_dst = 'e2:7a:09:79:0b:0f'
        dl_dst_bin = self.haddr_to_bin(dl_dst)
        mask = 'ff:ff:ff:ff:ff:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_DST, dp.ofproto.OXM_OF_ETH_DST_W]
        self._set_verify(headers, dl_dst_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_dst_masked(dl_dst_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_dst_masked_00(self, dp):
        dl_dst = 'e2:7a:09:79:0b:0f'
        dl_dst_bin = self.haddr_to_bin(dl_dst)
        mask = '00:00:00:00:00:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_DST, dp.ofproto.OXM_OF_ETH_DST_W]
        self._set_verify(headers, dl_dst_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_dst_masked(dl_dst_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_src(self, dp):
        dl_src = 'e2:7a:09:79:0b:0f'
        dl_src_bin = self.haddr_to_bin(dl_src)

        headers = [dp.ofproto.OXM_OF_ETH_SRC, dp.ofproto.OXM_OF_ETH_SRC_W]
        self._set_verify(headers, dl_src_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_src(dl_src_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_src_masked_ff(self, dp):
        dl_src = 'e2:7a:09:79:0b:0f'
        dl_src_bin = self.haddr_to_bin(dl_src)
        mask = 'ff:ff:ff:ff:ff:ff'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_SRC, dp.ofproto.OXM_OF_ETH_SRC_W]
        self._set_verify(headers, dl_src_bin, mask_bin, True, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_src_masked(dl_src_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_src_masked_f0(self, dp):
        dl_src = 'e2:7a:09:79:0b:0f'
        dl_src_bin = self.haddr_to_bin(dl_src)
        mask = 'ff:ff:ff:ff:ff:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_SRC, dp.ofproto.OXM_OF_ETH_SRC_W]
        self._set_verify(headers, dl_src_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_src_masked(dl_src_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_src_masked_00(self, dp):
        dl_src = 'e2:7a:09:79:0b:0f'
        dl_src_bin = self.haddr_to_bin(dl_src)
        mask = '00:00:00:00:00:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ETH_SRC, dp.ofproto.OXM_OF_ETH_SRC_W]
        self._set_verify(headers, dl_src_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_src_masked(dl_src_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_dl_type_ip(self, dp):
        dl_type = ether.ETH_TYPE_IP

        headers = [dp.ofproto.OXM_OF_ETH_TYPE]
        self._set_verify(headers, dl_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        self.add_matches(dp, match)

    def test_rule_set_dl_type_arp(self, dp):
        dl_type = ether.ETH_TYPE_ARP

        headers = [dp.ofproto.OXM_OF_ETH_TYPE]
        self._set_verify(headers, dl_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        self.add_matches(dp, match)

    def test_rule_set_dl_type_vlan(self, dp):
        dl_type = ether.ETH_TYPE_8021Q

        headers = [dp.ofproto.OXM_OF_ETH_TYPE]
        self._set_verify(headers, dl_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        self.add_matches(dp, match)

    def test_rule_set_dl_type_ipv6(self, dp):
        dl_type = ether.ETH_TYPE_IPV6

        headers = [dp.ofproto.OXM_OF_ETH_TYPE]
        self._set_verify(headers, dl_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        self.add_matches(dp, match)

    def test_rule_set_dl_type_lacp(self, dp):
        dl_type = ether.ETH_TYPE_SLOW

        headers = [dp.ofproto.OXM_OF_ETH_TYPE]
        self._set_verify(headers, dl_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        self.add_matches(dp, match)

    def test_rule_set_ip_dscp(self, dp):
        ip_dscp = 36
        dl_type = ether.ETH_TYPE_IP

        headers = [dp.ofproto.OXM_OF_IP_DSCP]
        self._set_verify(headers, ip_dscp)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_dscp(ip_dscp)
        self.add_matches(dp, match)

    def test_rule_set_vlan_vid(self, dp):
        vlan_vid = 0x4ef

        headers = [dp.ofproto.OXM_OF_VLAN_VID, dp.ofproto.OXM_OF_VLAN_VID_W]
        self._set_verify(headers, vlan_vid)

        match = dp.ofproto_parser.OFPMatch()
        match.set_vlan_vid(vlan_vid)
        self.add_matches(dp, match)

    def test_rule_set_vlan_vid_masked_ff(self, dp):
        vlan_vid = 0x4ef
        mask = 0xfff

        headers = [dp.ofproto.OXM_OF_VLAN_VID, dp.ofproto.OXM_OF_VLAN_VID_W]
        self._set_verify(headers, vlan_vid, mask, True)

        match = dp.ofproto_parser.OFPMatch()
        match.set_vlan_vid_masked(vlan_vid, mask)
        self.add_matches(dp, match)

    def test_rule_set_vlan_vid_masked_f0(self, dp):
        vlan_vid = 0x4ef
        mask = 0xff0

        headers = [dp.ofproto.OXM_OF_VLAN_VID, dp.ofproto.OXM_OF_VLAN_VID_W]
        self._set_verify(headers, vlan_vid, mask)

        match = dp.ofproto_parser.OFPMatch()
        match.set_vlan_vid_masked(vlan_vid, mask)
        self.add_matches(dp, match)

    def test_rule_set_vlan_vid_masked_00(self, dp):
        vlan_vid = 0x4ef
        mask = 0x000

        headers = [dp.ofproto.OXM_OF_VLAN_VID, dp.ofproto.OXM_OF_VLAN_VID_W]
        self._set_verify(headers, vlan_vid, mask)

        match = dp.ofproto_parser.OFPMatch()
        match.set_vlan_vid_masked(vlan_vid, mask)
        self.add_matches(dp, match)

    def test_rule_set_vlan_pcp(self, dp):
        vlan_vid = 0x4ef
        vlan_pcp = 5

        headers = [dp.ofproto.OXM_OF_VLAN_PCP]
        self._set_verify(headers, vlan_pcp)

        match = dp.ofproto_parser.OFPMatch()
        match.set_vlan_vid(vlan_vid)
        match.set_vlan_pcp(vlan_pcp)
        self.add_matches(dp, match)

    def test_rule_set_ip_ecn(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_ecn = 3

        headers = [dp.ofproto.OXM_OF_IP_ECN]
        self._set_verify(headers, ip_ecn)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_ecn(ip_ecn)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_icmp(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_ICMP

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_tcp(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_TCP

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_udp(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_UDP

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_ipv6_route(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ROUTING

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_ipv6_frag(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_FRAGMENT

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_ipv6_icmp(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_ipv6_none(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_NONE

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ip_proto_ipv6_dstopts(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_DSTOPTS

        headers = [dp.ofproto.OXM_OF_IP_PROTO]
        self._set_verify(headers, ip_proto)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_src(self, dp):
        dl_type = ether.ETH_TYPE_IP
        src = '192.168.196.250'
        src_int = self.ipv4_to_int(src)

        headers = [dp.ofproto.OXM_OF_IPV4_SRC, dp.ofproto.OXM_OF_IPV4_SRC_W]
        self._set_verify(headers, src_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_src(src_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_src_masked_32(self, dp):
        dl_type = ether.ETH_TYPE_IP
        src = '192.168.196.250'
        src_int = self.ipv4_to_int(src)
        mask = '255.255.255.255'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_SRC, dp.ofproto.OXM_OF_IPV4_SRC_W]
        self._set_verify(headers, src_int, mask_int, True, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_src_masked(src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_src_masked_24(self, dp):
        dl_type = ether.ETH_TYPE_IP
        src = '192.168.196.250'
        src_int = self.ipv4_to_int(src)
        mask = '255.255.255.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_SRC, dp.ofproto.OXM_OF_IPV4_SRC_W]
        self._set_verify(headers, src_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_src_masked(src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_src_masked_0(self, dp):
        dl_type = ether.ETH_TYPE_IP
        src = '192.168.196.250'
        src_int = self.ipv4_to_int(src)
        mask = '0.0.0.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_SRC, dp.ofproto.OXM_OF_IPV4_SRC_W]
        self._set_verify(headers, src_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_src_masked(src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_dst(self, dp):
        dl_type = ether.ETH_TYPE_IP
        dst = '192.168.54.155'
        dst_int = self.ipv4_to_int(dst)

        headers = [dp.ofproto.OXM_OF_IPV4_DST, dp.ofproto.OXM_OF_IPV4_DST_W]
        self._set_verify(headers, dst_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_dst(dst_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_dst_masked_32(self, dp):
        dl_type = ether.ETH_TYPE_IP
        dst = '192.168.54.155'
        dst_int = self.ipv4_to_int(dst)
        mask = '255.255.255.255'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_DST, dp.ofproto.OXM_OF_IPV4_DST_W]
        self._set_verify(headers, dst_int, mask_int, True, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_dst_masked(dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_dst_masked_24(self, dp):
        dl_type = ether.ETH_TYPE_IP
        dst = '192.168.54.155'
        dst_int = self.ipv4_to_int(dst)
        mask = '255.255.255.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_DST, dp.ofproto.OXM_OF_IPV4_DST_W]
        self._set_verify(headers, dst_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_dst_masked(dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv4_dst_masked_0(self, dp):
        dl_type = ether.ETH_TYPE_IP
        dst = '192.168.54.155'
        dst_int = self.ipv4_to_int(dst)
        mask = '0.0.0.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV4_DST, dp.ofproto.OXM_OF_IPV4_DST_W]
        self._set_verify(headers, dst_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv4_dst_masked(dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_tcp_src(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_TCP
        tp_src = 1103

        headers = [dp.ofproto.OXM_OF_TCP_SRC]
        self._set_verify(headers, tp_src)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_tcp_src(tp_src)
        self.add_matches(dp, match)

    def test_rule_set_tcp_dst(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_TCP
        tp_dst = 236

        headers = [dp.ofproto.OXM_OF_TCP_DST]
        self._set_verify(headers, tp_dst)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_tcp_dst(tp_dst)
        self.add_matches(dp, match)

    def test_rule_set_udp_src(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_UDP
        tp_src = 56617

        headers = [dp.ofproto.OXM_OF_UDP_SRC]
        self._set_verify(headers, tp_src)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_udp_src(tp_src)
        self.add_matches(dp, match)

    def test_rule_set_udp_dst(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_UDP
        tp_dst = 61278

        headers = [dp.ofproto.OXM_OF_UDP_DST]
        self._set_verify(headers, tp_dst)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_udp_dst(tp_dst)
        self.add_matches(dp, match)

    def test_rule_set_icmpv4_type(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_ICMP
        icmp_type = 8

        headers = [dp.ofproto.OXM_OF_ICMPV4_TYPE]
        self._set_verify(headers, icmp_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv4_type(icmp_type)
        self.add_matches(dp, match)

    def test_rule_set_icmpv4_code(self, dp):
        dl_type = ether.ETH_TYPE_IP
        ip_proto = inet.IPPROTO_ICMP
        icmp_type = 9
        icmp_code = 16

        headers = [dp.ofproto.OXM_OF_ICMPV4_CODE]
        self._set_verify(headers, icmp_code)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv4_type(icmp_type)
        match.set_icmpv4_code(icmp_code)
        self.add_matches(dp, match)

    def test_rule_set_arp_opcode(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_op = 1

        headers = [dp.ofproto.OXM_OF_ARP_OP]
        self._set_verify(headers, arp_op)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_opcode(arp_op)
        self.add_matches(dp, match)

    def test_rule_set_arp_spa(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_src = '192.168.222.57'
        nw_src_int = self.ipv4_to_int(nw_src)

        headers = [dp.ofproto.OXM_OF_ARP_SPA, dp.ofproto.OXM_OF_ARP_SPA_W]
        self._set_verify(headers, nw_src_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_spa(nw_src_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_spa_masked_32(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_src = '192.168.222.57'
        nw_src_int = self.ipv4_to_int(nw_src)
        mask = '255.255.255.255'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SPA, dp.ofproto.OXM_OF_ARP_SPA_W]
        self._set_verify(headers, nw_src_int, mask_int, True, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_spa_masked(nw_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_spa_masked_24(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_src = '192.168.222.57'
        nw_src_int = self.ipv4_to_int(nw_src)
        mask = '255.255.255.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SPA, dp.ofproto.OXM_OF_ARP_SPA_W]
        self._set_verify(headers, nw_src_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_spa_masked(nw_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_spa_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_src = '192.168.222.57'
        nw_src_int = self.ipv4_to_int(nw_src)
        mask = '0.0.0.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SPA, dp.ofproto.OXM_OF_ARP_SPA_W]
        self._set_verify(headers, nw_src_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_spa_masked(nw_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_tpa(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_dst = '192.168.198.233'
        nw_dst_int = self.ipv4_to_int(nw_dst)

        headers = [dp.ofproto.OXM_OF_ARP_TPA, dp.ofproto.OXM_OF_ARP_TPA_W]
        self._set_verify(headers, nw_dst_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tpa(nw_dst_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_tpa_masked_32(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_dst = '192.168.198.233'
        nw_dst_int = self.ipv4_to_int(nw_dst)
        mask = '255.255.255.255'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_TPA, dp.ofproto.OXM_OF_ARP_TPA_W]
        self._set_verify(headers, nw_dst_int, mask_int, True, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tpa_masked(nw_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_tpa_masked_24(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_dst = '192.168.198.233'
        nw_dst_int = self.ipv4_to_int(nw_dst)
        mask = '255.255.255.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_TPA, dp.ofproto.OXM_OF_ARP_TPA_W]
        self._set_verify(headers, nw_dst_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tpa_masked(nw_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_tpa_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        nw_dst = '192.168.198.233'
        nw_dst_int = self.ipv4_to_int(nw_dst)
        mask = '0.0.0.0'
        mask_int = self.ipv4_to_int(mask)

        headers = [dp.ofproto.OXM_OF_ARP_TPA, dp.ofproto.OXM_OF_ARP_TPA_W]
        self._set_verify(headers, nw_dst_int, mask_int, type_='ipv4')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tpa_masked(nw_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_arp_sha(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_sha = '3e:ec:13:9b:f3:0b'
        arp_sha_bin = self.haddr_to_bin(arp_sha)

        headers = [dp.ofproto.OXM_OF_ARP_SHA, dp.ofproto.OXM_OF_ARP_SHA_W]
        self._set_verify(headers, arp_sha_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_sha(arp_sha_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_sha_masked_ff(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_sha = '3e:ec:13:9b:f3:0b'
        arp_sha_bin = self.haddr_to_bin(arp_sha)
        mask = 'ff:ff:ff:ff:ff:ff'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SHA, dp.ofproto.OXM_OF_ARP_SHA_W]
        self._set_verify(headers, arp_sha_bin, mask_bin, True, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_sha_masked(arp_sha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_sha_masked_f0(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_sha = '3e:ec:13:9b:f3:0b'
        arp_sha_bin = self.haddr_to_bin(arp_sha)
        mask = 'ff:ff:ff:ff:ff:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SHA, dp.ofproto.OXM_OF_ARP_SHA_W]
        self._set_verify(headers, arp_sha_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_sha_masked(arp_sha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_sha_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_sha = '3e:ec:13:9b:f3:0b'
        arp_sha_bin = self.haddr_to_bin(arp_sha)
        mask = '00:00:00:00:00:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_SHA, dp.ofproto.OXM_OF_ARP_SHA_W]
        self._set_verify(headers, arp_sha_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_sha_masked(arp_sha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_tha(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_tha = '83:6c:21:52:49:68'
        arp_tha_bin = self.haddr_to_bin(arp_tha)

        headers = [dp.ofproto.OXM_OF_ARP_THA, dp.ofproto.OXM_OF_ARP_THA_W]
        self._set_verify(headers, arp_tha_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tha(arp_tha_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_tha_masked_ff(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_tha = '83:6c:21:52:49:68'
        arp_tha_bin = self.haddr_to_bin(arp_tha)
        mask = 'ff:ff:ff:ff:ff:ff'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_THA, dp.ofproto.OXM_OF_ARP_THA_W]
        self._set_verify(headers, arp_tha_bin, mask_bin, True, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tha_masked(arp_tha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_tha_masked_f0(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_tha = '83:6c:21:52:49:68'
        arp_tha_bin = self.haddr_to_bin(arp_tha)
        mask = 'ff:ff:ff:ff:ff:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_THA, dp.ofproto.OXM_OF_ARP_THA_W]
        self._set_verify(headers, arp_tha_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tha_masked(arp_tha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_arp_tha_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_ARP
        arp_tha = '83:6c:21:52:49:68'
        arp_tha_bin = self.haddr_to_bin(arp_tha)
        mask = '00:00:00:00:00:00'
        mask_bin = self.haddr_to_bin(mask)

        headers = [dp.ofproto.OXM_OF_ARP_THA, dp.ofproto.OXM_OF_ARP_THA_W]
        self._set_verify(headers, arp_tha_bin, mask_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_arp_tha_masked(arp_tha_bin, mask_bin)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_src(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_src = '2001:db8:bd05:1d2:288a:1fc0:1:10ee'
        ipv6_src_int = self.ipv6_to_int(ipv6_src)

        headers = [dp.ofproto.OXM_OF_IPV6_SRC, dp.ofproto.OXM_OF_IPV6_SRC_W]
        self._set_verify(headers, ipv6_src_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_src(ipv6_src_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_src_masked_ff(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_src = '2001:db8:bd05:1d2:288a:1fc0:1:10ee'
        ipv6_src_int = self.ipv6_to_int(ipv6_src)
        mask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_SRC, dp.ofproto.OXM_OF_IPV6_SRC_W]
        self._set_verify(headers, ipv6_src_int, mask_int, True, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_src_masked(ipv6_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_src_masked_f0(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_src = '2001:db8:bd05:1d2:288a:1fc0:1:10ee'
        ipv6_src_int = self.ipv6_to_int(ipv6_src)
        mask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:0'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_SRC, dp.ofproto.OXM_OF_IPV6_SRC_W]
        self._set_verify(headers, ipv6_src_int, mask_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_src_masked(ipv6_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_src_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_src = '2001:db8:bd05:1d2:288a:1fc0:1:10ee'
        ipv6_src_int = self.ipv6_to_int(ipv6_src)
        mask = '0:0:0:0:0:0:0:0'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_SRC, dp.ofproto.OXM_OF_IPV6_SRC_W]
        self._set_verify(headers, ipv6_src_int, mask_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_src_masked(ipv6_src_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_dst(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_dst = 'e9e8:9ea5:7d67:82cc:ca54:1fc0:2d24:f038'
        ipv6_dst_int = self.ipv6_to_int(ipv6_dst)

        headers = [dp.ofproto.OXM_OF_IPV6_DST, dp.ofproto.OXM_OF_IPV6_DST_W]
        self._set_verify(headers, ipv6_dst_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_dst(ipv6_dst_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_dst_masked_ff(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_dst = 'e9e8:9ea5:7d67:82cc:ca54:1fc0:2d24:f038'
        ipv6_dst_int = self.ipv6_to_int(ipv6_dst)
        mask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_DST, dp.ofproto.OXM_OF_IPV6_DST_W]
        self._set_verify(headers, ipv6_dst_int, mask_int, True, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_dst_masked(ipv6_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_dst_masked_f0(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_dst = 'e9e8:9ea5:7d67:82cc:ca54:1fc0:2d24:f038'
        ipv6_dst_int = self.ipv6_to_int(ipv6_dst)
        mask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:0'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_DST, dp.ofproto.OXM_OF_IPV6_DST_W]
        self._set_verify(headers, ipv6_dst_int, mask_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_dst_masked(ipv6_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_dst_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_dst = 'e9e8:9ea5:7d67:82cc:ca54:1fc0:2d24:f038'
        ipv6_dst_int = self.ipv6_to_int(ipv6_dst)
        mask = '0:0:0:0:0:0:0:0'
        mask_int = self.ipv6_to_int(mask)

        headers = [dp.ofproto.OXM_OF_IPV6_DST, dp.ofproto.OXM_OF_IPV6_DST_W]
        self._set_verify(headers, ipv6_dst_int, mask_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_dst_masked(ipv6_dst_int, mask_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_flabel(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_label = 0xc5384

        headers = [dp.ofproto.OXM_OF_IPV6_FLABEL,
                   dp.ofproto.OXM_OF_IPV6_FLABEL_W]
        self._set_verify(headers, ipv6_label)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_flabel(ipv6_label)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_flabel_masked_ff(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_label = 0xc5384
        mask = 0xfffff

        headers = [dp.ofproto.OXM_OF_IPV6_FLABEL,
                   dp.ofproto.OXM_OF_IPV6_FLABEL_W]
        self._set_verify(headers, ipv6_label, mask, True)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_flabel_masked(ipv6_label, mask)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_flabel_masked_f0(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_label = 0xc5384
        mask = 0xffff0

        headers = [dp.ofproto.OXM_OF_IPV6_FLABEL,
                   dp.ofproto.OXM_OF_IPV6_FLABEL_W]
        self._set_verify(headers, ipv6_label, mask)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_flabel_masked(ipv6_label, mask)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_flabel_masked_00(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ipv6_label = 0xc5384
        mask = 0x0

        headers = [dp.ofproto.OXM_OF_IPV6_FLABEL,
                   dp.ofproto.OXM_OF_IPV6_FLABEL_W]
        self._set_verify(headers, ipv6_label, mask)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ipv6_flabel_masked(ipv6_label, mask)
        self.add_matches(dp, match)

    def test_rule_set_icmpv6_type(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6
        icmp_type = 129

        headers = [dp.ofproto.OXM_OF_ICMPV6_TYPE]
        self._set_verify(headers, icmp_type)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv6_type(icmp_type)
        self.add_matches(dp, match)

    def test_rule_set_icmpv6_code(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6
        icmp_type = 138
        icmp_code = 1

        headers = [dp.ofproto.OXM_OF_ICMPV6_CODE]
        self._set_verify(headers, icmp_code)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv6_type(icmp_type)
        match.set_icmpv6_code(icmp_code)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_nd_target(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6
        icmp_type = 135
        target = "5420:db3f:921b:3e33:2791:98f:dd7f:2e19"
        target_int = self.ipv6_to_int(target)

        headers = [dp.ofproto.OXM_OF_IPV6_ND_TARGET]
        self._set_verify(headers, target_int, type_='ipv6')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv6_type(icmp_type)
        match.set_ipv6_nd_target(target_int)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_nd_sll(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6
        icmp_type = 135
        nd_sll = "93:6d:d0:d4:e8:36"
        nd_sll_bin = self.haddr_to_bin(nd_sll)

        headers = [dp.ofproto.OXM_OF_IPV6_ND_SLL]
        self._set_verify(headers, nd_sll_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv6_type(icmp_type)
        match.set_ipv6_nd_sll(nd_sll_bin)
        self.add_matches(dp, match)

    def test_rule_set_ipv6_nd_tll(self, dp):
        dl_type = ether.ETH_TYPE_IPV6
        ip_proto = inet.IPPROTO_ICMPV6
        icmp_type = 136
        nd_tll = "18:f6:66:b6:f1:b3"
        nd_tll_bin = self.haddr_to_bin(nd_tll)

        headers = [dp.ofproto.OXM_OF_IPV6_ND_TLL]
        self._set_verify(headers, nd_tll_bin, type_='mac')

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_ip_proto(ip_proto)
        match.set_icmpv6_type(icmp_type)
        match.set_ipv6_nd_tll(nd_tll_bin)
        self.add_matches(dp, match)

    def test_rule_set_mpls_label(self, dp):
        dl_type = 0x8847
        label = 2144

        headers = [dp.ofproto.OXM_OF_MPLS_LABEL]
        self._set_verify(headers, label)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_mpls_label(label)
        self.add_matches(dp, match)

    def test_rule_set_mpls_tc(self, dp):
        dl_type = 0x8847
        tc = 3

        headers = [dp.ofproto.OXM_OF_MPLS_TC]
        self._set_verify(headers, tc)

        match = dp.ofproto_parser.OFPMatch()
        match.set_dl_type(dl_type)
        match.set_mpls_tc(tc)
        self.add_matches(dp, match)

    def is_supported(self, t):
        # Open vSwitch 1.10 does not support MPLS yet.
        unsupported = [
            'test_rule_set_mpls_label',
            'test_rule_set_mpls_tc',
        ]
        for u in unsupported:
            if t.find(u) != -1:
                return False

        return True



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 BGP protocol implementation.
"""
import logging
import socket
import struct
import traceback
from socket import IPPROTO_TCP, TCP_NODELAY
from eventlet import semaphore

from os_ken.lib.packet import bgp
from os_ken.lib.packet.bgp import AS_TRANS
from os_ken.lib.packet.bgp import BGPMessage
from os_ken.lib.packet.bgp import BGPOpen
from os_ken.lib.packet.bgp import BGPUpdate
from os_ken.lib.packet.bgp import BGPKeepAlive
from os_ken.lib.packet.bgp import BGPNotification
from os_ken.lib.packet.bgp import BGP_MSG_OPEN
from os_ken.lib.packet.bgp import BGP_MSG_UPDATE
from os_ken.lib.packet.bgp import BGP_MSG_KEEPALIVE
from os_ken.lib.packet.bgp import BGP_MSG_NOTIFICATION
from os_ken.lib.packet.bgp import BGP_MSG_ROUTE_REFRESH
from os_ken.lib.packet.bgp import BGP_CAP_FOUR_OCTET_AS_NUMBER
from os_ken.lib.packet.bgp import BGP_CAP_ENHANCED_ROUTE_REFRESH
from os_ken.lib.packet.bgp import BGP_CAP_MULTIPROTOCOL
from os_ken.lib.packet.bgp import BGP_ERROR_HOLD_TIMER_EXPIRED
from os_ken.lib.packet.bgp import BGP_ERROR_SUB_HOLD_TIMER_EXPIRED
from os_ken.lib.packet.bgp import get_rf

from os_ken.services.protocols.bgp.base import Activity
from os_ken.services.protocols.bgp.base import add_bgp_error_metadata
from os_ken.services.protocols.bgp.base import BGPSException
from os_ken.services.protocols.bgp.base import CORE_ERROR_CODE
from os_ken.services.protocols.bgp.constants import BGP_FSM_CONNECT
from os_ken.services.protocols.bgp.constants import BGP_FSM_OPEN_CONFIRM
from os_ken.services.protocols.bgp.constants import BGP_FSM_OPEN_SENT
from os_ken.services.protocols.bgp.constants import BGP_VERSION_NUM
from os_ken.services.protocols.bgp.protocol import Protocol

LOG = logging.getLogger('bgpspeaker.speaker')

# BGP min. and max. message lengths as per RFC.
BGP_MIN_MSG_LEN = 19
BGP_MAX_MSG_LEN = 4096

# Keep-alive singleton.
_KEEP_ALIVE = BGPKeepAlive()


@add_bgp_error_metadata(code=CORE_ERROR_CODE, sub_code=2,
                        def_desc='Unknown error occurred related to Speaker.')
class BgpProtocolException(BGPSException):
    """Base exception related to peer connection management.
    """
    pass


def notification_factory(code, subcode):
    """Returns a `Notification` message corresponding to given codes.

    Parameters:
    - `code`: (int) BGP error code
    - `subcode`: (int) BGP error sub-code
    """
    notification = BGPNotification(code, subcode)
    if not notification.reason:
        raise ValueError('Invalid code/sub-code.')

    return notification


class BgpProtocol(Protocol, Activity):
    """Protocol that handles BGP messages.
    """
    MESSAGE_MARKER = (b'\xff\xff\xff\xff\xff\xff\xff\xff'
                      b'\xff\xff\xff\xff\xff\xff\xff\xff')

    def __init__(self, socket, signal_bus, is_reactive_conn=False):
        # Validate input.
        if socket is None:
            raise ValueError('Invalid arguments passed.')
        self._remotename = self.get_remotename(socket)
        self._localname = self.get_localname(socket)
        activity_name = ('BgpProtocol %s, %s, %s' % (is_reactive_conn,
                                                     self._remotename,
                                                     self._localname))
        Activity.__init__(self, name=activity_name)
        # Initialize instance variables.
        self._peer = None
        self._recv_buff = b''
        self._socket = socket
        self._socket.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
        self._sendlock = semaphore.Semaphore()
        self._signal_bus = signal_bus
        self._holdtime = None
        self._keepalive = None
        self._expiry = None
        # Add socket to Activity's socket container for managing it.
        if is_reactive_conn:
            self._asso_socket_map['passive_conn'] = self._socket
        else:
            self._asso_socket_map['active_conn'] = self._socket
        self._open_msg = None
        self.state = BGP_FSM_CONNECT
        self._is_reactive = is_reactive_conn
        self.sent_open_msg = None
        self.recv_open_msg = None
        self._is_bound = False
        self.cap_four_octet_as_number = False

    @property
    def is_reactive(self):
        return self._is_reactive

    @property
    def holdtime(self):
        return self._holdtime

    @property
    def keepalive(self):
        return self._keepalive

    def is_colliding(self, other_protocol):
        if not isinstance(other_protocol, BgpProtocol):
            raise ValueError('Currently only support comparing with '
                             '`BgpProtocol`')

        # Compare protocol connection end point's addresses
        if (self._remotename[0] == other_protocol._remotename[0] and
                self._localname[0] == other_protocol._localname[0]):
            return True

        return False

    def is_local_router_id_greater(self):
        """Compares *True* if local router id is greater when compared to peer
        bgp id.

        Should only be called after protocol has reached OpenConfirm state.
        """
        from os_ken.services.protocols.bgp.utils.bgp import from_inet_ptoi

        if not self.state == BGP_FSM_OPEN_CONFIRM:
            raise BgpProtocolException(desc='Can access remote router id only'
                                            ' after open message is received')
        remote_id = self.recv_open_msg.bgp_identifier
        local_id = self.sent_open_msg.bgp_identifier
        return from_inet_ptoi(local_id) > from_inet_ptoi(remote_id)

    def is_enhanced_rr_cap_valid(self):
        """Checks is enhanced route refresh capability is enabled/valid.

        Checks sent and received `Open` messages to see if this session with
        peer is capable of enhanced route refresh capability.
        """
        if not self.recv_open_msg:
            raise ValueError('Did not yet receive peers open message.')

        err_cap_enabled = False
        local_caps = self.sent_open_msg.opt_param
        peer_caps = self.recv_open_msg.opt_param

        local_cap = [cap for cap in local_caps
                     if cap.cap_code == BGP_CAP_ENHANCED_ROUTE_REFRESH]
        peer_cap = [cap for cap in peer_caps
                    if cap.cap_code == BGP_CAP_ENHANCED_ROUTE_REFRESH]

        # Both local and peer should advertise ERR capability for it to be
        # enabled.
        if local_cap and peer_cap:
            err_cap_enabled = True

        return err_cap_enabled

    def _check_route_fmly_adv(self, open_msg, route_family):
        match_found = False

        local_caps = open_msg.opt_param
        for cap in local_caps:
            # Check MP_BGP capability was advertised.
            if cap.cap_code == BGP_CAP_MULTIPROTOCOL:
                # Iterate over all advertised mp_bgp caps to find a match.
                if (route_family.afi == cap.afi and
                        route_family.safi == cap.safi):
                    match_found = True

        return match_found

    def is_route_family_adv(self, route_family):
        """Checks if `route_family` was advertised to peer as per MP_BGP cap.

        Returns:
            - True: if given address family was advertised.
            - False: if given address family was not advertised.
        """
        return self._check_route_fmly_adv(self.sent_open_msg, route_family)

    def is_route_family_adv_recv(self, route_family):
        """Checks if `route_family` was advertised by peer as per MP_BGP cap.

        Returns:
            - True: if given address family was advertised.
            - False: if given address family was not advertised.
        """
        return self._check_route_fmly_adv(self.recv_open_msg, route_family)

    @property
    def negotiated_afs(self):
        local_caps = self.sent_open_msg.opt_param
        remote_caps = self.recv_open_msg.opt_param

        local_mbgp_cap = [cap for cap in local_caps
                          if cap.cap_code == BGP_CAP_MULTIPROTOCOL]
        remote_mbgp_cap = [cap for cap in remote_caps
                           if cap.cap_code == BGP_CAP_MULTIPROTOCOL]

        # Check MP_BGP capabilities were advertised.
        if local_mbgp_cap and remote_mbgp_cap:
            local_families = set([
                (peer_cap.afi, peer_cap.safi)
                for peer_cap in local_mbgp_cap
            ])
            remote_families = set([
                (peer_cap.afi, peer_cap.safi)
                for peer_cap in remote_mbgp_cap
            ])
            afi_safi = local_families.intersection(remote_families)
        else:
            afi_safi = set()

        afs = []
        for afi, safi in afi_safi:
            afs.append(get_rf(afi, safi))
        return afs

    def is_mbgp_cap_valid(self, route_family):
        """Returns True if both sides of this protocol have advertise
        capability for this address family.
        """
        return (self.is_route_family_adv(route_family) and
                self.is_route_family_adv_recv(route_family))

    def is_four_octet_as_number_cap_valid(self):
        """Returns True if both sides of this protocol have Four-Octet
        AS number capability."""
        return (self.cap_four_octet_as_number and
                self._peer.cap_four_octet_as_number)

    def _run(self, peer):
        """Sends open message to peer and handles received messages.

        Parameters:
            - `peer`: the peer to which this protocol instance is connected to.
        """
        # We know the peer we are connected to, we send open message.
        self._peer = peer
        self.connection_made()

        # We wait for peer to send messages.
        self._recv_loop()

    def data_received(self, next_bytes):
        try:
            self._data_received(next_bytes)
        except bgp.BgpExc as exc:
            LOG.error(
                "BGPExc Exception while receiving data: "
                "%s \n Traceback %s \n"
                % (str(exc), traceback.format_exc())
            )
            if exc.SEND_ERROR:
                self.send_notification(exc.CODE, exc.SUB_CODE)
            else:
                self._socket.close()
            raise exc

    @staticmethod
    def parse_msg_header(buff):
        """Parses given `buff` into bgp message header format.

        Returns a tuple of marker, length, type of bgp message.
        """
        return struct.unpack('!16sHB', buff)

    def _data_received(self, next_bytes):
        """Maintains buffer of bytes received from peer and extracts bgp
        message from this buffer if enough data is received.

        Validates bgp message marker, length, type and data and constructs
        appropriate bgp message instance and calls handler.

        :Parameters:
            - `next_bytes`: next set of bytes received from peer.
        """
        # Append buffer with received bytes.
        self._recv_buff += next_bytes

        while True:
            # If current buffer size is less then minimum bgp message size, we
            # return as we do not have a complete bgp message to work with.
            if len(self._recv_buff) < BGP_MIN_MSG_LEN:
                return

            # Parse message header into elements.
            auth, length, ptype = BgpProtocol.parse_msg_header(
                self._recv_buff[:BGP_MIN_MSG_LEN])

            # Check if we have valid bgp message marker.
            # We should get default marker since we are not supporting any
            # authentication.
            if (auth != BgpProtocol.MESSAGE_MARKER):
                LOG.error('Invalid message marker received: %s', auth)
                raise bgp.NotSync()

            # Check if we have valid bgp message length.
            check = (length < BGP_MIN_MSG_LEN or length > BGP_MAX_MSG_LEN)

            # RFC says: The minimum length of the OPEN message is 29
            # octets (including the message header).
            check2 = (ptype == BGP_MSG_OPEN and length < BGPOpen._MIN_LEN)

            # RFC says: A KEEPALIVE message consists of only the
            # message header and has a length of 19 octets.
            check3 = (ptype == BGP_MSG_KEEPALIVE and
                      length != BGPKeepAlive._MIN_LEN)

            # RFC says: The minimum length of the UPDATE message is 23
            # octets.
            check4 = (ptype == BGP_MSG_UPDATE and
                      length < BGPUpdate._MIN_LEN)

            if any((check, check2, check3, check4)):
                raise bgp.BadLen(ptype, length)

            # If we have partial message we wait for rest of the message.
            if len(self._recv_buff) < length:
                return
            msg, _, rest = BGPMessage.parser(self._recv_buff)
            self._recv_buff = rest

            # If we have a valid bgp message we call message handler.
            self._handle_msg(msg)

    def send_notification(self, code, subcode):
        """Utility to send notification message.

        Closes the socket after sending the message.
        :Parameters:
            - `socket`: (socket) - socket over which to send notification
             message.
            - `code`: (int) - BGP Notification code
            - `subcode`: (int) - BGP Notification sub-code

        RFC ref: http://tools.ietf.org/html/rfc4486
        http://www.iana.org/assignments/bgp-parameters/bgp-parameters.xhtml
        """
        notification = BGPNotification(code, subcode)
        reason = notification.reason
        self._send_with_lock(notification)
        self._signal_bus.bgp_error(self._peer, code, subcode, reason)
        if len(self._localname):
            LOG.error('Sent notification to %r >> %s', self._localname,
                      notification)
        self._socket.close()

    def _send_with_lock(self, msg):
        self._sendlock.acquire()
        try:
            self._socket.sendall(msg.serialize())
        except socket.error:
            self.connection_lost('failed to write to socket')
        finally:
            self._sendlock.release()

    def send(self, msg):
        if not self.started:
            raise BgpProtocolException('Tried to send message to peer when '
                                       'this protocol instance is not started'
                                       ' or is no longer is started state.')
        self._send_with_lock(msg)

        if msg.type == BGP_MSG_NOTIFICATION:
            LOG.error('Sent notification to %s >> %s', self._remotename, msg)

            self._signal_bus.bgp_notification_sent(self._peer, msg)
        else:
            LOG.debug('Sent msg to %s >> %s', self._remotename, msg)

    def stop(self):
        Activity.stop(self)

    def _validate_open_msg(self, open_msg):
        """Validates BGP OPEN message according from application context.

        Parsing modules takes care of validating OPEN message that need no
        context. But here we validate it according to current application
        settings. RTC or RR/ERR are MUST capability if peer does not support
        either one of them we have to end session.
        """
        assert open_msg.type == BGP_MSG_OPEN

        opt_param_cap_map = open_msg.opt_param_cap_map

        # Validate remote AS number.
        remote_as = open_msg.my_as
        # Try to get AS number from Four-Octet AS number capability.
        cap4as = opt_param_cap_map.get(BGP_CAP_FOUR_OCTET_AS_NUMBER, None)
        if cap4as is None:
            if remote_as == AS_TRANS:
                # Raise Bad Peer AS error message, if my_as is AS_TRANS
                # and without Four-Octet AS number capability.
                raise bgp.BadPeerAs()
            self.cap_four_octet_as_number = False
        else:
            # Note: Even if the peer has Four-Octet AS number capability,
            # keep the local capability setting
            remote_as = cap4as.as_number
            self.cap_four_octet_as_number = True
        #  Validate remote AS number with local setting.
        if remote_as != self._peer.remote_as:
            raise bgp.BadPeerAs()

        # Validate bgp version number.
        if open_msg.version != BGP_VERSION_NUM:
            raise bgp.UnsupportedVersion(BGP_VERSION_NUM)

    def _handle_msg(self, msg):
        """When a BGP message is received, send it to peer.

        Open messages are validated here. Peer handler is called to handle each
        message except for *Open* and *Notification* message. On receiving
        *Notification* message we close connection with peer.
        """
        LOG.debug('Received msg from %s << %s', self._remotename, msg)

        # If we receive open message we try to bind to protocol
        if msg.type == BGP_MSG_OPEN:
            if self.state == BGP_FSM_OPEN_SENT:
                # Validate open message.
                self._validate_open_msg(msg)
                self.recv_open_msg = msg
                self.state = BGP_FSM_OPEN_CONFIRM
                self._peer.state.bgp_state = self.state

                # Try to bind this protocol to peer.
                self._is_bound = self._peer.bind_protocol(self)

                # If this protocol failed to bind to peer.
                if not self._is_bound:
                    # Failure to bind to peer indicates connection collision
                    # resolution choose different instance of protocol and this
                    # instance has to close. Before closing it sends
                    # appropriate notification msg. to peer.
                    raise bgp.CollisionResolution()

                # If peer sends Hold Time as zero, then according to RFC we do
                # not set Hold Time and Keep Alive timer.
                if msg.hold_time == 0:
                    LOG.info('The Hold Time sent by the peer is zero, hence '
                             'not setting any Hold Time and Keep Alive'
                             ' timers.')
                else:
                    # Start Keep Alive timer considering Hold Time preference
                    # of the peer.
                    self._start_timers(msg.hold_time)
                    self._send_keepalive()

                # Peer does not see open message.
                return
            else:
                # If we receive a Open message out of order
                LOG.error('Open message received when current state is not '
                          'OpenSent')
                # Received out-of-order open message
                # We raise Finite state machine error
                raise bgp.FiniteStateMachineError()
        elif msg.type == BGP_MSG_NOTIFICATION:
            if self._peer:
                self._signal_bus.bgp_notification_received(self._peer, msg)
            # If we receive notification message
            LOG.error('Received notification message, hence closing '
                      'connection %s', msg)
            self._socket.close()
            return

        # If we receive keepalive or update message, we reset expire timer.
        if (msg.type == BGP_MSG_KEEPALIVE or
                msg.type == BGP_MSG_UPDATE):
            if self._expiry:
                self._expiry.reset()

        # Call peer message handler for appropriate messages.
        if (msg.type in
                (BGP_MSG_UPDATE, BGP_MSG_KEEPALIVE, BGP_MSG_ROUTE_REFRESH)):
            self._peer.handle_msg(msg)
        # We give chance to other threads to run.
        self.pause(0)

    def _start_timers(self, peer_holdtime):
        """Starts keepalive and expire timers.

        Hold time is set to min. of peer and configured/default hold time.
        Starts keep alive timer and expire timer based on this value.
        """
        neg_timer = min(self._holdtime, peer_holdtime)
        if neg_timer < self._holdtime:
            LOG.info('Negotiated hold time (%s) is lower then '
                     'configured/default (%s).', neg_timer, self._holdtime)
        # We use negotiated timer value.
        self._holdtime = neg_timer
        self._keepalive = self._create_timer('Keepalive Timer',
                                             self._send_keepalive)
        interval = self._holdtime // 3
        self._keepalive.start(interval, now=False)
        # Setup the expire timer.
        self._expiry = self._create_timer('Holdtime Timer', self._expired)
        self._expiry.start(self._holdtime, now=False)
        LOG.debug('Started keep-alive and expire timer for negotiated hold'
                  'time %s', self._holdtime)

    def _expired(self):
        """Hold timer expired event handler.
        """
        LOG.info('Negotiated hold time %s expired.', self._holdtime)
        code = BGP_ERROR_HOLD_TIMER_EXPIRED
        subcode = BGP_ERROR_SUB_HOLD_TIMER_EXPIRED
        self.send_notification(code, subcode)
        self.connection_lost('Negotiated hold time %s expired.' %
                             self._holdtime)
        self.stop()

    def _send_keepalive(self):
        self.send(_KEEP_ALIVE)

    def _recv_loop(self):
        """Sits in tight loop collecting data received from peer and
        processing it.
        """
        required_len = BGP_MIN_MSG_LEN
        conn_lost_reason = "Connection lost as protocol is no longer active"
        try:
            while True:
                next_bytes = self._socket.recv(required_len)
                if len(next_bytes) == 0:
                    conn_lost_reason = 'Peer closed connection'
                    break
                self.data_received(next_bytes)
        except socket.error as err:
            conn_lost_reason = 'Connection to peer lost: %s.' % err
        except bgp.BgpExc as ex:
            conn_lost_reason = 'Connection to peer lost, reason: %s.' % ex
        except Exception as e:
            LOG.debug(traceback.format_exc())
            conn_lost_reason = str(e)
        finally:
            self.connection_lost(conn_lost_reason)

    def connection_made(self):
        """Connection to peer handler.

        We send bgp open message to peer and initialize related attributes.
        """
        assert self.state == BGP_FSM_CONNECT
        # We have a connection with peer we send open message.
        open_msg = self._peer.create_open_msg()
        self._holdtime = open_msg.hold_time
        self.state = BGP_FSM_OPEN_SENT
        if not self.is_reactive:
            self._peer.state.bgp_state = self.state
        self.sent_open_msg = open_msg
        self.send(open_msg)
        self._peer.connection_made()

    def connection_lost(self, reason):
        """Stops all timers and notifies peer that connection is lost.
        """

        if self._peer:
            state = self._peer.state.bgp_state
            if self._is_bound or state == BGP_FSM_OPEN_SENT:
                self._peer.connection_lost(reason)

            self._peer = None

        if reason:
            LOG.info(reason)
        else:
            LOG.info('Connection to peer closed for unknown reasons.')



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import unittest
import logging
import struct
from struct import *
from nose.tools import *
from os_ken.ofproto import ether, inet
from os_ken.lib.packet.packet import Packet
from os_ken.lib.packet.udp import udp
from os_ken.lib.packet.ipv4 import ipv4
from os_ken.lib.packet import packet_utils
from os_ken.lib import addrconv


LOG = logging.getLogger('test_udp')


class Test_udp(unittest.TestCase):
    """ Test case for udp
    """
    src_port = 6431
    dst_port = 8080
    total_length = 65507
    csum = 12345
    u = udp(src_port, dst_port, total_length, csum)
    buf = pack(udp._PACK_STR, src_port, dst_port, total_length, csum)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.src_port, self.u.src_port)
        eq_(self.dst_port, self.u.dst_port)
        eq_(self.total_length, self.u.total_length)
        eq_(self.csum, self.u.csum)

    def test_parser(self):
        r1, r2, _ = self.u.parser(self.buf)

        eq_(self.src_port, r1.src_port)
        eq_(self.dst_port, r1.dst_port)
        eq_(self.total_length, r1.total_length)
        eq_(self.csum, r1.csum)
        eq_(None, r2)

    def test_serialize(self):
        src_port = 6431
        dst_port = 8080
        total_length = 0
        csum = 0

        src_ip = '192.168.10.1'
        dst_ip = '192.168.100.1'
        prev = ipv4(4, 5, 0, 0, 0, 0, 0, 64,
                    inet.IPPROTO_UDP, 0, src_ip, dst_ip)

        u = udp(src_port, dst_port, total_length, csum)
        buf = u.serialize(bytearray(), prev)
        res = struct.unpack(udp._PACK_STR, buf)

        eq_(res[0], src_port)
        eq_(res[1], dst_port)
        eq_(res[2], struct.calcsize(udp._PACK_STR))

        # checksum
        ph = struct.pack('!4s4sBBH',
                         addrconv.ipv4.text_to_bin(src_ip),
                         addrconv.ipv4.text_to_bin(dst_ip), 0, 17, res[2])
        d = ph + buf + bytearray()
        s = packet_utils.checksum(d)
        eq_(0, s)

    @raises(Exception)
    def test_malformed_udp(self):
        m_short_buf = self.buf[1:udp._MIN_LEN]
        udp.parser(m_short_buf)

    def test_default_args(self):
        prev = ipv4(proto=inet.IPPROTO_UDP)
        u = udp()
        buf = u.serialize(bytearray(), prev)
        res = struct.unpack(udp._PACK_STR, buf)

        eq_(res[0], 1)
        eq_(res[1], 1)
        eq_(res[2], udp._MIN_LEN)

    def test_json(self):
        jsondict = self.u.to_jsondict()
        u = udp.from_jsondict(jsondict['udp'])
        eq_(str(self.u), str(u))



# Copyright (C) 2011 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2011 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class EventBase(object):
    """
    The base of all event classes.

    A OSKen application can define its own event type by creating a subclass.
    """

    def __init__(self):
        super(EventBase, self).__init__()


class EventRequestBase(EventBase):
    """
    The base class for synchronous request for OSKenApp.send_request.
    """

    def __init__(self):
        super(EventRequestBase, self).__init__()
        self.dst = None  # app.name of provide the event.
        self.src = None
        self.sync = False
        self.reply_q = None


class EventReplyBase(EventBase):
    """
    The base class for synchronous request reply for OSKenApp.send_reply.
    """

    def __init__(self, dst):
        super(EventReplyBase, self).__init__()
        self.dst = dst



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
 Module that holds various constants.

 This module helps in breaking circular dependencies too.
"""

# Various states of bgp state machine.
BGP_FSM_IDLE = 'Idle'
BGP_FSM_CONNECT = 'Connect'
BGP_FSM_ACTIVE = 'Active'
BGP_FSM_OPEN_SENT = 'OpenSent'
BGP_FSM_OPEN_CONFIRM = 'OpenConfirm'
BGP_FSM_ESTABLISHED = 'Established'

# All valid BGP finite state machine states.
BGP_FSM_VALID_STATES = (BGP_FSM_IDLE, BGP_FSM_CONNECT, BGP_FSM_ACTIVE,
                        BGP_FSM_OPEN_SENT, BGP_FSM_OPEN_CONFIRM,
                        BGP_FSM_ESTABLISHED)

# Supported bgp protocol version number.
BGP_VERSION_NUM = 4

# Standard BGP server port number.
STD_BGP_SERVER_PORT_NUM = 179

#
# Constants used to indicate VRF prefix source.
#
# It indicates prefix inside VRF table came from bgp peer to VPN table and then
# to VRF table..
VPN_TABLE = 'vpn_table'
VRF_TABLE = 'vrf_table'

# RTC EOR timer default value
# Time to wait for RTC-EOR, before we can send initial UPDATE as per RFC
RTC_EOR_DEFAULT_TIME = 60

# Constants for AttributeMaps
ATTR_MAPS_ORG_KEY = '__orig'
ATTR_MAPS_LABEL_KEY = 'at_maps_key'
ATTR_MAPS_LABEL_DEFAULT = 'default'
ATTR_MAPS_VALUE = 'at_maps'
